import datetime as dt
from typing import ClassVar, Optional, cast

from cuenca_validations.types import (
    SavingCategory,
    SavingRequest,
    SavingUpdateRequest,
    StrictPositiveInt,
    WalletQuery,
)

from .base import Updateable, Wallet


class Saving(Wallet, Updateable):
    _resource: ClassVar = 'savings'
    _query_params: ClassVar = WalletQuery
    name: str
    category: SavingCategory
    goal_amount: Optional[StrictPositiveInt]
    goal_date: Optional[dt.datetime]

    @classmethod
    def create(
        cls,
        name: str,
        category: SavingCategory,
        goal_amount: Optional[int] = None,
        goal_date: Optional[dt.datetime] = None,
    ):
        request = SavingRequest(
            name=name,
            category=category,
            goal_amount=goal_amount,
            goal_date=goal_date,
        )
        return cast('Saving', cls._create(**request.dict()))

    @classmethod
    def update(
        cls,
        saving_id: str,
        name: Optional[str] = None,
        category: Optional[SavingCategory] = None,
        goal_amount: Optional[int] = None,
        goal_date: Optional[dt.datetime] = None,
    ) -> 'Saving':
        request = SavingUpdateRequest(
            name=name,
            category=category,
            goal_amount=goal_amount,
            goal_date=goal_date,
        )
        return cast('Saving', cls._update(id=saving_id, **request.dict()))



import requests_mock  # type: ignore

import cuenca


def test_get_balance():
    # It is the case when the user has transactions in the account
    response_json = {
        'items': [
            {
                'id': 'LE6xvD1ocHmSIu7MgzlC9woj',
                'created_at': '2021-08-17T21:45:23.061000',
                'user_id': 'US3zGWi1n852bTyqD1wCZ1ft',
                'name': 'RESTAURANT BRASIL MEXICO DF CMXMX',
                'amount': 1212,
                'descriptor': '',
                'rolling_balance': 7578889,
                'type': 'debit',
                'related_transaction_uri': '/card_transactions/CT3TA',
                'funding_instrument_uri': '/cards/CA243qFpGLDAaPamCMYHhS0p',
                'wallet_id': 'default',
            }
        ],
        'next_page_uri': None,
    }
    with requests_mock.mock() as m:
        m.get(
            'https://sandbox.cuenca.com/balance_entries?limit=1',
            status_code=200,
            json=response_json,
        )

        balance = cuenca.get_balance()
    assert balance == response_json['items'][0]['rolling_balance']


def test_get_balance_before_first_transaction():
    # When the user have no transactions at all
    # balance_entries endpoint returns `items` as empty list.
    # It means that its balance is Mx$0.00
    response_json = {'items': [], 'next_page_uri': None}

    with requests_mock.mock() as m:
        m.get(
            'https://sandbox.cuenca.com/balance_entries?limit=1',
            status_code=200,
            json=response_json,
        )

        balance = cuenca.get_balance()
    assert balance == 0



import pytest

from cuenca import CardTransaction


@pytest.mark.vcr
def test_card_transaction_parent_retrieve():
    ct = CardTransaction.retrieve('CT17YWLDHWI59r6F0oG1ihvK')
    related_cts = ct.related_card_transactions
    assert len(related_cts) == 2
    for card_transaction in related_cts:
        assert card_transaction.type != 'auth'


@pytest.mark.vcr
def test_card_transaction_child_retrieve():
    ct = CardTransaction.retrieve('CT4jC3Dh65FNmb0lGBTnvr4I')
    related_cts = ct.related_card_transactions
    assert len(related_cts) == 1
    assert related_cts[0].type == 'auth'


@pytest.mark.vcr
def test_card_transaction_refund_without_parent_id_retrieve():
    ct = CardTransaction.retrieve('CTKLh6VNCRTybvfpVSMLu8m')
    assert not ct.related_card_transactions


@pytest.mark.vcr
def test_card_transaction_expiration_retrieve():
    ct = CardTransaction.retrieve('CT4kMz2YNoWxVSQOCPCh2nQD')
    assert ct.type == 'expiration'


@pytest.mark.vcr
def test_card_transaction_retrieve_card():
    ct = CardTransaction.retrieve('CT00PARENT01')
    card = ct.card
    assert card
    assert card.user_id == ct.user_id
    assert len(card.number) == 16



import datetime as dt
from typing import ClassVar, Optional, cast

from cuenca_validations.types import (
    Address,
    CurpField,
    KYCVerificationUpdateRequest,
    Rfc,
)

from ..http import Session, session as global_session
from .base import Creatable, Retrievable, Updateable


class KYCVerification(Creatable, Retrievable, Updateable):
    _resource: ClassVar = 'kyc_verifications'

    platform_id: str
    created_at: dt.datetime
    deactivated_at: Optional[dt.datetime]
    verification_id: Optional[str]
    curp: Optional[CurpField] = None
    rfc: Optional[Rfc] = None
    address: Optional[Address] = None

    @classmethod
    def create(cls, session: Session = global_session) -> 'KYCVerification':
        return cast('KYCVerification', cls._create(session=session))

    @classmethod
    def update(
        cls,
        kyc_id: str,
        curp: Optional[CurpField] = None,
    ) -> 'KYCVerification':
        req = KYCVerificationUpdateRequest(curp=curp)
        return cast('KYCVerification', cls._update(id=kyc_id, **req.dict()))



import datetime as dt
from unittest.mock import MagicMock

import pytest

from cuenca.exc import MalformedJwtToken
from cuenca.jwt import Jwt

TEST_TOKEN = (
    'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6'
    'IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyLCJleHAiOjE2MDYyNjI0MDB9.gbrIYP6Q0yf2'
    'V3GEgxnzO5fNHKgHPqaIzZ-cjvFlnik'
)


@pytest.mark.parametrize(
    'expires_at, should_be_expired',
    [
        (dt.datetime.utcnow() + dt.timedelta(days=-1), True),
        (dt.datetime.utcnow() + dt.timedelta(days=1), False),
    ],
)
def test_expired(expires_at: dt.datetime, should_be_expired: bool):
    jwt = Jwt(expires_at=expires_at, token='_')
    assert jwt.is_expired == should_be_expired


def test_get_expiration_date():
    assert Jwt.get_expiration_date(TEST_TOKEN) == dt.datetime(2020, 11, 25)


def test_invalid_token():
    test_token = 'INVALID'
    with pytest.raises(MalformedJwtToken):
        Jwt.get_expiration_date(test_token)


def test_create_token():
    session = MagicMock()
    session.post.return_value = dict(token=TEST_TOKEN)
    jwt = Jwt.create(session)
    assert jwt.expires_at == dt.datetime(2020, 11, 25)
    assert jwt.token == TEST_TOKEN



import pytest

from cuenca import Deposit


@pytest.mark.vcr
def test_deposit_retrieve():
    id_deposit = 'SP01'
    deposit: Deposit = Deposit.retrieve(id_deposit)
    assert deposit.id == id_deposit
    account = deposit.source
    assert account is not None



import datetime as dt
from typing import ClassVar, Optional, cast

from cuenca_validations.types import TransferNetwork

from .accounts import Account
from .base import Transaction
from .resources import retrieve_uri


class WhatsappTransfer(Transaction):
    _resource: ClassVar = 'whatsapp_transfers'

    updated_at: dt.datetime
    recipient_name: str
    phone_number: str
    claim_url: Optional[str]
    expires_at: dt.datetime
    # defined after the transfer has been claimed
    destination_uri: Optional[str]
    network: Optional[TransferNetwork]
    tracking_key: Optional[str]  # clave rastreo if network is SPEI

    @property  # type: ignore
    def destination(self) -> Optional[Account]:
        if self.destination_uri is None:
            dest = None
        else:
            dest = cast(Account, retrieve_uri(self.destination_uri))
        return dest



import datetime as dt
from io import BytesIO
from typing import Dict

import pytest
from cuenca_validations.types import Country, Gender, State

import cuenca

cuenca.configure(sandbox=True)


@pytest.fixture(scope='module')
def vcr_config():
    config = dict()
    config['filter_headers'] = [
        ('Authorization', 'DUMMY'),
        ('X-Cuenca-Token', 'DUMMY'),
    ]
    return config


@pytest.fixture
def transfer():
    yield cuenca.Transfer.create(
        account_number='646180157034181180',
        amount=10000,
        descriptor='first transfer',
        recipient_name='Frida Kahlo',
    )


@pytest.fixture
def curp_validation_request() -> Dict:
    curp_validation = dict(
        names='José',
        first_surname='López',
        second_surname='Hernández',
        date_of_birth=dt.date(1966, 6, 6),
        state_of_birth=State.DF,
        country_of_birth=Country.MX,
        gender=Gender.male,
    )
    return curp_validation


@pytest.fixture
def user_request() -> Dict:
    user_dict = dict(
        curp='LOHJ660606HDFPRS02',
        phone_number='+525511223344',
        email_address='jose@test.com',
        profession='employee',
        address=dict(
            street='calle 1',
            ext_number='2',
            int_number='3',
            postal_code='09900',
            state=State.DF.value,
            country=Country.MX,
        ),
    )
    return user_dict


@pytest.fixture
def file() -> BytesIO:
    with open('tests/data/test_file.jpeg', 'rb') as image_file:
        return BytesIO(image_file.read())



from typing import ClassVar, cast

from ..http import Session, session as global_session
from .base import Creatable


class LoginToken(Creatable):
    _resource: ClassVar = 'login_tokens'

    @classmethod
    def create(cls, session: Session = global_session) -> 'LoginToken':
        """
        Use this method to create a token that will last for 7 days
        Make sure to store this token in a safe place
        :return: Token that you can use in cuenca.configure
        """
        return cast('LoginToken', cls._create(session=session))



import pytest

from cuenca.resources import CurpValidation


@pytest.mark.vcr
def test_curp_validations_create(curp_validation_request) -> None:
    curp_validation = CurpValidation.create(**curp_validation_request)
    assert curp_validation.id is not None
    assert curp_validation.renapo_curp_match


@pytest.mark.vcr
def test_curp_validations_retrieve():
    curp_validation_id = 'CVhY5Pi6yBQJO22p1lg-_EPw'
    curp_validation = CurpValidation.retrieve(curp_validation_id)
    assert curp_validation.id == curp_validation_id



import pytest

from cuenca.resources import CurpValidation, Identity


@pytest.mark.vcr
def test_identity_retrieve(curp_validation_request):
    # creating a curp_validation automatically creates the identity
    curp_validation = CurpValidation.create(**curp_validation_request)
    assert curp_validation.renapo_curp_match
    # querying the identity
    identity = Identity.one(curp=curp_validation.calculated_curp)
    assert identity.id is not None



import pytest

from cuenca import KYCVerification


@pytest.mark.vcr
def test_kyc_verification_create():
    kyc_verification: KYCVerification = KYCVerification.create()
    assert kyc_verification.id


@pytest.mark.vcr
def test_kyc_verification_retrieve():
    kyc_verification = KYCVerification.retrieve('KYC01')
    assert kyc_verification.id


@pytest.mark.vcr
def test_kyc_verification_update():
    kyc_id = 'KYC01'
    changes = dict(curp='HEMA921130HNERNN05')
    kyc_verification = KYCVerification.update(kyc_id, **changes)
    assert all(
        item in kyc_verification.to_dict().items() for item in changes.items()
    )



from typing import ClassVar, Dict, List, cast

from cuenca_validations.types import BatchFileMetadata, FileBatchUploadRequest

from ..http import Session, session as global_session
from .base import Creatable, Queryable


class FileBatch(Creatable, Queryable):
    _resource: ClassVar = 'file_batches'

    received_files: List[BatchFileMetadata]
    uploaded_files: List[BatchFileMetadata]
    user_id: str

    @classmethod
    def create(
        cls,
        files: List[Dict],
        user_id: str,
        *,
        session: Session = global_session,
    ) -> 'FileBatch':
        req = FileBatchUploadRequest(files=files, user_id=user_id)
        return cast('FileBatch', cls._create(session=session, **req.dict()))



import pytest

from cuenca import LoginToken, Transfer, UserLogin
from cuenca.http.client import Session


@pytest.fixture
def session():
    session = Session()
    session.configure(
        'api_key',
        'api_secret',
        sandbox=True,
    )
    return session


@pytest.mark.vcr
def test_login_token(session):
    UserLogin.create('222222', session=session)
    login_token = LoginToken.create(session=session)
    session.session.headers.pop('X-Cuenca-LoginId')
    session.configure(login_token=login_token.id)
    Transfer.count(session=session)



import datetime as dt
from typing import ClassVar, Optional, cast

from cuenca_validations.types.requests import CardActivationRequest

from ..http import Session, session as global_session
from .base import Creatable
from .cards import Card
from .resources import retrieve_uri


class CardActivation(Creatable):
    _resource: ClassVar = 'card_activations'

    created_at: dt.datetime
    user_id: str
    ip_address: str
    card_uri: Optional[str]
    success: bool

    @classmethod
    def create(
        cls,
        number: str,
        exp_month: int,
        exp_year: int,
        cvv2: str,
        *,
        session: Session = global_session,
    ) -> 'CardActivation':
        """
        Associates a physical card with the current user

        :param number: Card number
        :param exp_month:
        :param exp_year:
        :param cvv2:
        """
        req = CardActivationRequest(
            number=number,
            exp_month=exp_month,
            exp_year=exp_year,
            cvv2=cvv2,
        )
        return cast(
            'CardActivation', cls._create(session=session, **req.dict())
        )

    @property
    def card(self) -> Optional[Card]:
        result = None
        if self.card_uri:
            result = cast(Card, retrieve_uri(self.card_uri))
        return result



from typing import ClassVar, List, Optional, cast

from cuenca_validations.types import (
    CardErrorType,
    CardNetwork,
    CardTransactionQuery,
    CardTransactionType,
    CardType,
)

from .base import Transaction
from .cards import Card
from .resources import retrieve_uri, retrieve_uris


class CardTransaction(Transaction):
    _resource: ClassVar = 'card_transactions'
    _query_params: ClassVar = CardTransactionQuery

    type: CardTransactionType
    network: CardNetwork
    related_card_transaction_uris: List[str]
    card_uri: str
    card_last4: str
    card_type: CardType
    metadata: dict
    error_type: Optional[CardErrorType]

    @property  # type: ignore
    def related_card_transactions(self) -> Optional[List['CardTransaction']]:
        if not self.related_card_transaction_uris:
            return []
        return cast(
            List['CardTransaction'],
            retrieve_uris(self.related_card_transaction_uris),
        )

    @property  # type: ignore
    def card(self) -> Card:
        return cast(Card, retrieve_uri(self.card_uri))



import datetime as dt
from typing import ClassVar, Optional, cast

from cuenca_validations.types import SessionRequest, SessionType
from pydantic import AnyUrl

from .. import http
from .base import Creatable, Queryable, Retrievable


class Session(Creatable, Retrievable, Queryable):
    _resource: ClassVar = 'sessions'

    id: str
    created_at: dt.datetime
    user_id: str
    platform_id: str
    expires_at: dt.datetime
    success_url: Optional[AnyUrl]
    failure_url: Optional[AnyUrl]
    type: Optional[SessionType]

    @classmethod
    def create(
        cls,
        user_id: str,
        type: SessionType,
        success_url: Optional[str] = None,
        failure_url: Optional[str] = None,
        *,
        session: http.Session = http.session,
    ) -> 'Session':
        req = SessionRequest(
            user_id=user_id,
            type=type,
            success_url=success_url,
            failure_url=failure_url,
        )
        return cast('Session', cls._create(session=session, **req.dict()))



from typing import Dict

import pytest
from cuenca_validations.types import SessionType
from pydantic import ValidationError

import cuenca
from cuenca.resources import CurpValidation, Session, User


@pytest.mark.vcr
def test_session_create(curp_validation_request: Dict, user_request: Dict):
    curp_valdation = CurpValidation.create(**curp_validation_request)
    user_request['curp'] = curp_valdation.validated_curp
    user = User.create(**user_request)

    success_url = 'no url'
    with pytest.raises(ValidationError):
        user_session = Session.create(
            user.id,
            SessionType.registration,
            success_url=success_url,
        )

    success_url = 'https://example.com/succeeded_registration'
    failure_url = 'https://example.com/failed_registration'

    user_session = Session.create(
        user.id,
        SessionType.registration,
        success_url=success_url,
        failure_url=failure_url,
    )

    assert user_session.user_id == user.id
    assert user_session.type == SessionType.registration
    assert user_session.success_url == success_url
    assert user_session.failure_url == failure_url

    ephimeral_cuenca_session = cuenca.http.Session()
    ephimeral_cuenca_session.configure(session_token=user_session.id)
    user = User.update(user.id, email_address='manu@example.com')
    assert user.email_address == 'manu@example.com'



import base64
import datetime as dt
import json
from io import BytesIO
from typing import ClassVar, Dict, Generator, Optional, Union
from urllib.parse import urlencode

from cuenca_validations.types import (
    FileFormat,
    QueryParams,
    SantizedDict,
    TransactionQuery,
    TransactionStatus,
)
from pydantic import BaseModel

from ..exc import MultipleResultsFound, NoResultFound
from ..http import Session, session as global_session


class Resource(BaseModel):
    _resource: ClassVar[str]

    id: str

    @classmethod
    def _from_dict(cls, obj_dict: Dict[str, Union[str, int]]) -> 'Resource':
        cls._filter_excess_fields(obj_dict)
        return cls(**obj_dict)

    @classmethod
    def _filter_excess_fields(cls, obj_dict):
        """
        dataclasses don't allow __init__ to be called with excess fields. This
        method allows the API to add fields in the response body without
        breaking the client
        """
        excess = set(obj_dict.keys()) - set(
            cls.schema().get("properties").keys()
        )
        for f in excess:
            del obj_dict[f]

    def to_dict(self):
        return SantizedDict(self.dict())


class Retrievable(Resource):
    @classmethod
    def retrieve(
        cls, id: str, *, session: Session = global_session
    ) -> Resource:
        resp = session.get(f'/{cls._resource}/{id}')
        return cls._from_dict(resp)

    def refresh(self, *, session: Session = global_session):
        new = self.retrieve(self.id, session=session)
        for attr, value in new.__dict__.items():
            setattr(self, attr, value)


class Creatable(Resource):
    @classmethod
    def _create(cls, *, session: Session = global_session, **data) -> Resource:
        resp = session.post(cls._resource, data)
        return cls._from_dict(resp)


class Updateable(Resource):

    updated_at: dt.datetime

    @classmethod
    def _update(
        cls, id: str, *, session: Session = global_session, **data
    ) -> Resource:
        resp = session.patch(f'/{cls._resource}/{id}', data)
        return cls._from_dict(resp)


class Deactivable(Resource):
    deactivated_at: Optional[dt.datetime]

    @classmethod
    def deactivate(
        cls, id: str, *, session: Session = global_session, **data
    ) -> Resource:
        resp = session.delete(f'/{cls._resource}/{id}', data)
        return cls._from_dict(resp)

    @property
    def is_active(self):
        return not self.deactivated_at


class Downloadable(Resource):
    @classmethod
    def download(
        cls,
        instance,
        file_format: FileFormat,
        *,
        session: Session = global_session,
    ) -> BytesIO:
        resp = session.request(
            'get',
            f'/{cls._resource}/{instance.id}',
            headers=dict(Accept=file_format.value),
        )
        return BytesIO(resp)

    @property
    def pdf(self) -> bytes:
        return self.download(self, file_format=FileFormat.pdf).read()

    @property
    def xml(self) -> bytes:
        return self.download(self, file_format=FileFormat.xml).read()


class Uploadable(Resource):
    @classmethod
    def _upload(
        cls,
        file: bytes,
        user_id: str,
        *,
        session: Session = global_session,
        **data,
    ) -> Resource:
        encoded_file = base64.b64encode(file)
        resp = session.request(
            'post',
            cls._resource,
            files=dict(
                file=(None, encoded_file),
                user_id=(None, user_id),
                **{k: (None, v) for k, v in data.items()},
            ),
        )
        return cls._from_dict(json.loads(resp))


class Queryable(Resource):
    _query_params: ClassVar = QueryParams

    created_at: dt.datetime

    @classmethod
    def one(
        cls, *, session: Session = global_session, **query_params
    ) -> Resource:
        q = cls._query_params(limit=2, **query_params)
        resp = session.get(cls._resource, q.dict())
        items = resp['items']
        len_items = len(items)
        if not len_items:
            raise NoResultFound
        if len_items > 1:
            raise MultipleResultsFound
        return cls._from_dict(items[0])

    @classmethod
    def first(
        cls, *, session: Session = global_session, **query_params
    ) -> Optional[Resource]:
        q = cls._query_params(limit=1, **query_params)
        resp = session.get(cls._resource, q.dict())
        try:
            item = resp['items'][0]
        except IndexError:
            rv = None
        else:
            rv = cls._from_dict(item)
        return rv

    @classmethod
    def count(
        cls, *, session: Session = global_session, **query_params
    ) -> int:
        q = cls._query_params(count=True, **query_params)
        resp = session.get(cls._resource, q.dict())
        return resp['count']

    @classmethod
    def all(
        cls, *, session: Session = global_session, **query_params
    ) -> Generator[Resource, None, None]:
        session = session or global_session
        q = cls._query_params(**query_params)
        next_page_uri = f'{cls._resource}?{urlencode(q.dict())}'
        while next_page_uri:
            page = session.get(next_page_uri)
            yield from (cls._from_dict(item) for item in page['items'])
            next_page_uri = page['next_page_uri']


class Transaction(Retrievable, Queryable):
    _query_params: ClassVar = TransactionQuery

    user_id: str
    amount: int  # in centavos
    status: TransactionStatus
    descriptor: str  # how it appears for the customer


class Wallet(Creatable, Deactivable, Retrievable, Queryable):
    user_id: str
    balance: int

    @property
    def wallet_uri(self):
        return f'/{self._resource}/{self.id}'



__all__ = [
    '__version__',
    'ApiKey',
    'Account',
    'Arpc',
    'BalanceEntry',
    'BillPayment',
    'Card',
    'CardActivation',
    'CardTransaction',
    'CardValidation',
    'CurpValidation',
    'Commission',
    'Deposit',
    'Endpoint',
    'File',
    'FileBatch',
    'Identity',
    'IdentityEvent',
    'KYCVerification',
    'LimitedWallet',
    'LoginToken',
    'Saving',
    'ServiceProvider',
    'Session',
    'Statement',
    'Transfer',
    'User',
    'UserCredential',
    'UserEvent',
    'UserLogin',
    'Verification',
    'WalletTransaction',
    'Webhook',
    'WhatsappTransfer',
    'configure',
    'get_balance',
]

from typing import cast

from . import http
from .resources import (
    Account,
    ApiKey,
    Arpc,
    BalanceEntry,
    BillPayment,
    Card,
    CardActivation,
    CardTransaction,
    CardValidation,
    Commission,
    CurpValidation,
    Deposit,
    Endpoint,
    File,
    FileBatch,
    Identity,
    IdentityEvent,
    KYCVerification,
    LimitedWallet,
    LoginToken,
    Saving,
    ServiceProvider,
    Session,
    Statement,
    Transfer,
    User,
    UserCredential,
    UserEvent,
    UserLogin,
    Verification,
    WalletTransaction,
    Webhook,
    WhatsappTransfer,
)
from .version import __version__

configure = http.session.configure
session = http.session


def get_balance(session: http.Session = session) -> int:
    balance_entry = cast('BalanceEntry', BalanceEntry.first(session=session))
    return balance_entry.rolling_balance if balance_entry else 0



import pytest
from cuenca_validations.types import VerificationType

from cuenca import Verification
from cuenca.exc import CuencaResponseException


@pytest.mark.vcr
def test_verification_email_create():
    verification: Verification = Verification.create(
        recipient='mail@cuenca.com',
        type=VerificationType.email,
        platform_id='PL01',
    )
    assert verification.id


@pytest.mark.vcr
def test_verification_phone_create():
    verification: Verification = Verification.create(
        recipient='+525555555555',
        type=VerificationType.phone,
        platform_id='PL01',
    )
    assert verification.id


@pytest.mark.vcr
def test_verification_verify():
    verification = Verification.verify(id='VE02', code='299566')
    assert verification.id


@pytest.mark.vcr
def test_verification_verify_fail():
    with pytest.raises(CuencaResponseException) as exc:
        Verification.verify(id='VE01', code='222222')
        assert exc.status_code == 400
        assert exc.json['error'] == 'Invalid code.'


@pytest.mark.vcr
def test_verification_verify_fail_max_attempt():
    with pytest.raises(CuencaResponseException) as exc:
        Verification.verify(id='VE01', code='333333')
        assert exc.status_code == 400
        assert exc.json['error'] == 'Max Retries reached.'


@pytest.mark.vcr
def test_verification_verify_deactivated_verification():
    with pytest.raises(CuencaResponseException) as exc:
        Verification.verify(id='VE01', code='333333')
        assert exc.status_code == 403
        assert exc.json['error'] == (
            'You can not validate a deactivated Verification.'
        )



from typing import ClassVar, Optional, cast

from clabe import Clabe
from cuenca_validations.types import (
    AccountQuery,
    CurpField,
    LimitedWalletRequest,
    Rfc,
)

from .base import Wallet


class LimitedWallet(Wallet):
    _resource: ClassVar = 'limited_wallets'
    _query_params: ClassVar = AccountQuery
    account_number: Clabe
    allowed_rfc: Rfc
    allowed_curp: CurpField

    @classmethod
    def create(
        cls,
        allowed_curp: Optional[CurpField] = None,
        allowed_rfc: Optional[Rfc] = None,
    ) -> 'LimitedWallet':
        """
        Limited wallet is a special sub-account to receive money only from
        specific person, SPEI Deposits will be accepted only if the sender
        account curp/rfc match with registered data

        Args:
            allowed_curp: Valid CURP
            allowed_rfc: Valid RFC
        """
        request = LimitedWalletRequest(
            allowed_curp=allowed_curp,
            allowed_rfc=allowed_rfc,
        )
        return cast('LimitedWallet', cls._create(**request.dict()))



from typing import ClassVar

from cuenca_validations.types import AccountQuery

from .base import Queryable, Retrievable


class Account(Retrievable, Queryable):
    _resource: ClassVar = 'accounts'
    _query_params: ClassVar = AccountQuery

    name: str  # legal name provided by institution
    account_number: str
    institution_name: str



from typing import ClassVar, Optional, cast

from cuenca_validations.types import (
    WalletTransactionQuery,
    WalletTransactionRequest,
    WalletTransactionType,
)

from cuenca.resources.base import Creatable, Transaction
from cuenca.resources.resources import retrieve_uri

from .base import Wallet


class WalletTransaction(Transaction, Creatable):
    _resource: ClassVar = 'wallet_transactions'
    _query_params: ClassVar = WalletTransactionQuery

    transaction_type: WalletTransactionType
    wallet_uri: str

    @property
    def wallet(self) -> Optional['Wallet']:
        return cast('Wallet', retrieve_uri(self.wallet_uri))

    @classmethod
    def create(
        cls,
        wallet_uri: str,
        transaction_type: WalletTransactionType,
        amount: int,
    ):
        request = WalletTransactionRequest(
            wallet_uri=wallet_uri,
            transaction_type=transaction_type,
            amount=amount,
        )
        return cast('WalletTransaction', cls._create(**request.dict()))



import datetime as dt
from typing import ClassVar

from cuenca_validations.types import EventQuery, EventType

from .base import Queryable, Retrievable


class IdentityEvent(Retrievable, Queryable):
    _resource: ClassVar = 'identity_events'
    _query_params: ClassVar = EventQuery

    identity_id: str
    new_model: dict
    type: EventType
    created_at: dt.datetime



from typing import ClassVar, Optional, cast

from cuenca_validations.types import (
    CardFundingType,
    CardIssuer,
    CardStatus,
    CardType,
)
from cuenca_validations.types.queries import CardQuery
from cuenca_validations.types.requests import CardRequest, CardUpdateRequest

from cuenca.resources.base import Creatable, Queryable, Retrievable, Updateable

from ..http import Session, session as global_session

MAX_PIN_ATTEMPTS = 3


class Card(Retrievable, Queryable, Creatable, Updateable):
    _resource: ClassVar = 'cards'
    _query_params: ClassVar = CardQuery

    user_id: Optional[str]
    number: str
    exp_month: int
    exp_year: int
    cvv2: str
    pin: Optional[str]
    type: CardType
    status: CardStatus
    issuer: CardIssuer
    funding_type: CardFundingType
    pin_attempts_failed: Optional[int] = None

    @property
    def last_4_digits(self):
        return self.number[-4:]

    @property
    def bin(self):
        return self.number[:6]

    @property
    def pin_attempts_exceeded(self) -> bool:
        return (
            self.pin_attempts_failed >= MAX_PIN_ATTEMPTS
            if self.pin_attempts_failed
            else False
        )

    @classmethod
    def create(
        cls,
        issuer: CardIssuer,
        funding_type: CardFundingType,
        user_id: str = 'me',
        *,
        session: Session = global_session,
    ) -> 'Card':
        """
        Assigns user_id and ledger_account_id to a existing virtual card

        :param user_id: associated user id
        :param funding_type: debit or credit
        :param issuer:
        :return: New assigned card
        """
        req = CardRequest(
            user_id=user_id,
            issuer=issuer,
            funding_type=funding_type,
        )
        return cast('Card', cls._create(session=session, **req.dict()))

    @classmethod
    def update(
        cls,
        card_id: str,
        status: Optional[CardStatus] = None,
        pin_block: Optional[str] = None,
        *,
        session: Session = global_session,
    ) -> 'Card':
        """
        Updates card properties that are not sensitive or fixed data. It allows
        reconfigure properties like status, and manufacturer.

        :param card_id: existing card_id
        :param status:
        :param pin_block
        :param session:
        :return: Updated card object
        """
        req = CardUpdateRequest(status=status, pin_block=pin_block)
        resp = cls._update(card_id, session=session, **req.dict())
        return cast('Card', resp)

    @classmethod
    def deactivate(
        cls, card_id: str, *, session: Session = global_session
    ) -> 'Card':
        """
        Deactivates a card
        """
        url = f'{cls._resource}/{card_id}'
        resp = session.delete(url)
        return cast('Card', cls._from_dict(resp))



import pytest

from cuenca.resources import CardValidation


@pytest.mark.vcr
def test_card_validations():
    card_data = dict(
        number='5448750129965637',
        cvv='685',
        cvv2='150',
        icvv='399',
        exp_month=2,
        exp_year=25,
        pin_block='3B241739AB05D290',
        pin_attempts_exceeded=True,
    )
    validation = CardValidation.create(**card_data)
    assert validation.is_active
    assert validation.card_uri is not None
    assert validation.is_valid_cvv
    assert validation.is_valid_cvv2
    assert validation.is_valid_icvv
    assert validation.is_valid_pin_block
    assert validation.is_valid_exp_date
    assert not validation.is_expired
    c = validation.card
    assert validation.card_id == c.id
    assert validation.is_active
    assert validation.is_pin_attempts_exceeded



import pytest

from cuenca import Account


@pytest.mark.vcr
def test_query_account(transfer):
    account: Account = Account.one(account_number=transfer.account_number)
    destination = transfer.destination
    assert account.id == destination.id
    assert account.account_number == destination.account_number



import datetime as dt
from typing import ClassVar, Optional

from cuenca_validations.types import (
    Address,
    Gender,
    IdentityQuery,
    KYCFile,
    State,
    TOSAgreement,
    UserStatus,
    VerificationStatus,
)
from cuenca_validations.types.identities import CurpField

from .base import Queryable, Retrievable


class Identity(Retrievable, Queryable):
    _resource: ClassVar = 'identities'
    _query_params: ClassVar = IdentityQuery

    created_at: dt.datetime
    names: str
    first_surname: str
    second_surname: Optional[str]
    curp: Optional[CurpField]
    rfc: Optional[str]
    gender: Gender
    date_of_birth: Optional[dt.date]
    state_of_birth: Optional[State]
    country_of_birth: Optional[str]
    status: Optional[UserStatus]
    tos_agreement: Optional[TOSAgreement]
    blacklist_validation_status: Optional[VerificationStatus]
    address: Optional[Address]
    govt_id: Optional[KYCFile]
    proof_of_address: Optional[KYCFile]
    proof_of_life: Optional[KYCFile]



from typing import ClassVar, cast

from cuenca_validations.types import BillPaymentQuery

from .base import Transaction
from .resources import retrieve_uri
from .service_providers import ServiceProvider


class BillPayment(Transaction):
    _resource: ClassVar = 'bill_payments'
    _query_params: ClassVar = BillPaymentQuery

    account_number: str
    provider_uri: str

    @property
    def provider(self) -> ServiceProvider:
        provider = cast(ServiceProvider, retrieve_uri(self.provider_uri))
        return provider



import pytest

from cuenca import LimitedWallet


@pytest.mark.vcr
def test_limited_wallet_create():
    curp = 'TAXM840916HNEMXT02'
    rfc = 'TAXM840916123'
    wallet = LimitedWallet.create(allowed_curp=curp, allowed_rfc=rfc)
    assert wallet.id is not None
    assert wallet.balance == 0
    assert wallet.allowed_curp == curp
    assert wallet.allowed_rfc == rfc


@pytest.mark.vcr
def test_limited_wallet_retrieve():
    id = 'LA3vx08KlQCXsaP9qyWfb680'
    wallet = LimitedWallet.retrieve(id)
    assert wallet.id == id



import pytest
from cuenca_validations.types import KYCFileType

from cuenca import FileBatch


@pytest.mark.vcr
def test_file_batch_create():
    files = [
        dict(
            url='https://somefile.com/location=XXX1',
            type=KYCFileType.ine,
            is_back=True,
        ),
        dict(
            url='https://somefile.com/location=XXX2',
            type=KYCFileType.proof_of_liveness,
        ),
        dict(
            url='https://somefile.com/location=XXX3',
            type=KYCFileType.proof_of_address,
        ),
    ]
    user_id = 'US01'
    batch: FileBatch = FileBatch.create(files=files, user_id=user_id)
    assert batch.id is not None
    assert batch.received_files is not None
    assert batch.user_id == 'US01'
    assert len(batch.uploaded_files) == 3

    for file in batch.uploaded_files:
        assert file.id is not None
        assert file.url is not None
        if file.type == KYCFileType.ine:
            assert file.is_back



import datetime as dt
from typing import ClassVar, Optional, cast

from cuenca_validations.types.requests import UserCredentialRequest

from ..http import Session, session as global_session
from .base import Creatable


class UserLogin(Creatable):
    _resource: ClassVar = 'user_logins'

    last_login_at: Optional[dt.datetime]
    success: bool

    @classmethod
    def create(
        cls, password: str, *, session: Session = global_session
    ) -> 'UserLogin':
        req = UserCredentialRequest(password=password)
        login = cast('UserLogin', cls._create(session=session, **req.dict()))
        if login.success:
            session.session.headers['X-Cuenca-LoginId'] = login.id
        return login

    @classmethod
    def logout(
        cls, user_id: str = 'me', *, session: Session = global_session
    ) -> None:
        # Using user_id vs user_login_id to avoid needing to store
        # user_login_id or perform a query to fetch it
        session.delete(f'{cls._resource}/{user_id}', dict())
        session.session.headers.pop('X-Cuenca-LoginId', None)



from typing import Any, ClassVar, Dict

from cuenca_validations.types.enums import WebhookEvent

from .base import Queryable, Retrievable


class Webhook(Retrievable, Queryable):
    _resource: ClassVar = 'webhooks'

    payload: Dict[str, Any]
    event: WebhookEvent



import datetime as dt
from typing import ClassVar, List, Optional, cast

from cuenca_validations.types import (
    Address,
    AddressUpdateRequest,
    Beneficiary,
    KYCFile,
    KYCFileUpdateRequest,
    PhoneNumber,
    TOSAgreement,
    TOSUpdateRequest,
    UserQuery,
    UserRequest,
    UserStatus,
    UserUpdateRequest,
)
from cuenca_validations.types.identities import CurpField
from pydantic import EmailStr

from ..http import Session, session as global_session
from .base import Creatable, Queryable, Retrievable, Updateable
from .identities import Identity
from .resources import retrieve_uri


class User(Creatable, Retrievable, Updateable, Queryable):
    _resource: ClassVar = 'users'
    _query_params: ClassVar = UserQuery

    identity_uri: str
    level: int
    created_at: dt.datetime
    phone_number: Optional[PhoneNumber]
    email_address: Optional[EmailStr]
    profession: Optional[str]
    terms_of_service: Optional[TOSAgreement]
    status: Optional[UserStatus]
    address: Optional[Address]
    govt_id: Optional[KYCFile]
    proof_of_address: Optional[KYCFile]
    proof_of_life: Optional[KYCFile]
    beneficiaries: Optional[List[Beneficiary]]
    platform_id: Optional[str] = None

    @classmethod
    def create(
        cls,
        curp: CurpField,
        phone_number: Optional[PhoneNumber] = None,
        email_address: Optional[EmailStr] = None,
        profession: Optional[str] = None,
        address: Optional[Address] = None,
        email_verification_id: Optional[str] = None,
        phone_verification_id: Optional[str] = None,
        *,
        session: Session = global_session,
    ) -> 'User':
        req = UserRequest(
            curp=curp,
            phone_number=phone_number,
            email_address=email_address,
            profession=profession,
            address=address,
            email_verification_id=email_verification_id,
            phone_verification_id=phone_verification_id,
        )
        return cast('User', cls._create(session=session, **req.dict()))

    @classmethod
    def update(
        cls,
        user_id: str,
        phone_number: Optional[PhoneNumber] = None,
        email_address: Optional[str] = None,
        profession: Optional[str] = None,
        address: Optional[AddressUpdateRequest] = None,
        beneficiaries: Optional[List[Beneficiary]] = None,
        govt_id: Optional[KYCFileUpdateRequest] = None,
        proof_of_address: Optional[KYCFileUpdateRequest] = None,
        proof_of_life: Optional[KYCFileUpdateRequest] = None,
        terms_of_service: Optional[TOSUpdateRequest] = None,
        verification_id: Optional[str] = None,
        *,
        session: Session = global_session,
    ):
        request = UserUpdateRequest(
            phone_number=phone_number,
            email_address=email_address,
            profession=profession,
            address=address,
            beneficiaries=beneficiaries,
            govt_id=govt_id,
            proof_of_address=proof_of_address,
            proof_of_life=proof_of_life,
            terms_of_service=terms_of_service,
            verification_id=verification_id,
        )
        return cast(
            'User',
            cls._update(id=user_id, **request.dict(), session=session),
        )

    @property
    def identity(self) -> Identity:
        return cast(Identity, retrieve_uri(self.identity_uri))



import datetime as dt
from typing import ClassVar, Optional, Union, cast

from cuenca_validations.types import (
    VerificationAttemptRequest,
    VerificationRequest,
    VerificationType,
)
from cuenca_validations.types.identities import PhoneNumber
from pydantic import EmailStr

from ..http import Session, session as global_session
from .base import Creatable, Updateable


class Verification(Creatable, Updateable):
    _resource: ClassVar = 'verifications'

    recipient: Union[EmailStr, PhoneNumber]
    type: VerificationType
    created_at: dt.datetime
    deactivated_at: Optional[dt.datetime]

    @classmethod
    def create(
        cls,
        recipient: str,
        type: VerificationType,
        platform_id: str,
        session: Session = global_session,
    ) -> 'Verification':
        req = VerificationRequest(
            recipient=recipient, type=type, platform_id=platform_id
        )
        return cast('Verification', cls._create(**req.dict(), session=session))

    @classmethod
    def verify(
        cls,
        id: str,
        code: str,
        session: Session = global_session,
    ) -> 'Verification':
        req = VerificationAttemptRequest(code=code)
        return cast(
            'Verification',
            cls._update(id=id, **req.dict(), session=session),
        )



import re
from concurrent.futures import ThreadPoolExecutor
from typing import Dict, List, cast

from .base import Retrievable

ENDPOINT_RE = re.compile(r'.*/(?P<resource>[a-z_]+)/(?P<id>.+)$')
RESOURCES: Dict[str, Retrievable] = {}  # set in ./__init__.py after imports


def retrieve_uri(uri: str) -> Retrievable:
    m = ENDPOINT_RE.match(uri)
    if not m:
        raise ValueError(f'uri is not a valid format: {uri}')
    resource, id_ = m.groups()
    return cast(Retrievable, RESOURCES[resource].retrieve(id_))


def retrieve_uris(uris: List[str]) -> List[Retrievable]:
    with ThreadPoolExecutor(max_workers=len(uris)) as executor:
        return [obj for obj in executor.map(retrieve_uri, uris)]



from typing import ClassVar, List

from cuenca_validations.types import ServiceProviderCategory

from .base import Queryable, Retrievable


class ServiceProvider(Retrievable, Queryable):
    _resource: ClassVar = 'service_providers'

    name: str
    provider_key: str
    categories: List[ServiceProviderCategory]
